using SYSCore.ReportingService;
using System;
using System.Collections.Generic;
using System.Configuration;
using System.Drawing;
using System.Drawing.Imaging;
using System.IO;
using System.Linq;
using System.Net;
using System.ServiceModel;
using System.Threading.Tasks;
using System.Web;

namespace SYSCore.aControl
{
    public class ReportServices
    {
        public static async Task<string> GenerateReport(string reportPath, string reportType, List<ParameterValue> paramList, string fileName, string outputPath = null)
        {
            reportType = reportType.ToUpper();

            if (reportType != "EXCELOPENXML" && reportType != "PDF")
            {
                throw new System.ArgumentException("Unsupported file type.");
            }
            if (outputPath == null)
            {
                //Default output location
                outputPath = ConfigurationManager.AppSettings["DefaultReportOutputPath"];
            }

            try
            {
                ParameterValue[] parameters = paramList.ToArray();
                var render = RenderReportAsync(reportPath, reportType, parameters);
                await Task.WhenAll(render);

                if (render.Result != null)
                {
                    outputPath = System.Web.HttpContext.Current.Server.MapPath(outputPath);
                    string fileLocation = String.Concat(outputPath, fileName);

                    //Create Directory If not exists
                    FileInfo filePath = new FileInfo(outputPath);
                    filePath.Directory.Create();

                    //Save file to Directory
                    File.WriteAllBytes(fileLocation, render.Result);

                    return "Done";
                }
                else
                {
                    throw new System.ArgumentException("Something went wrong, please try again");
                }
            }
            catch (Exception ex)
            {
                throw (ex);
            }
        }

        public static async Task<byte[]> GenerateReportAsBytes(string reportPath, string reportType, List<ParameterValue> paramList, string fileName, string outputPath = null)
        {
            reportType = reportType.ToUpper();

            if (reportType != "EXCELOPENXML" && reportType != "PDF")
            {
                throw new System.ArgumentException("Unsupported file type.");
            }
            if (outputPath == null)
            {
                //Default output location
                outputPath = ConfigurationManager.AppSettings["DefaultReportOutputPath"];
            }

            try
            {
                ParameterValue[] parameters = paramList.ToArray();
                var render = RenderReportAsync(reportPath, reportType, parameters);
                await Task.WhenAll(render);

                if (render.Result != null)
                {
                    return render.Result;
                }
                else
                {
                    throw new System.ArgumentException("Something went wrong, please try again");
                }
            }
            catch (Exception ex)
            {
                throw (ex);
            }
        }

        public static async Task<ReportImageResult> RenderReportAsImage(string reportPath, List<ParameterValue> paramList, ReportOptions options = null)
        {

            ParameterValue[] parameters = paramList.ToArray();

            string reportEndpoint = ConfigurationManager.AppSettings["ReportEndpoint"];
            string reportEndpointUsername = ConfigurationManager.AppSettings["ReportUsername"];
            string reportEndpointPassword = ConfigurationManager.AppSettings["ReportPassword"];

            reportPath = String.Concat("/PMS_Rep/", reportPath);

            TimeSpan SetTimeOut = new TimeSpan(0, 30, 0);
            BasicHttpBinding rsBinding = new BasicHttpBinding();
            rsBinding.Security.Mode = BasicHttpSecurityMode.TransportCredentialOnly;
            rsBinding.Security.Transport.ClientCredentialType = HttpClientCredentialType.Ntlm;
            rsBinding.MaxReceivedMessageSize = Int32.MaxValue;
            rsBinding.MaxBufferSize = Int32.MaxValue;
            rsBinding.OpenTimeout = SetTimeOut;
            rsBinding.CloseTimeout = SetTimeOut;
            rsBinding.SendTimeout = SetTimeOut;
            rsBinding.ReceiveTimeout = SetTimeOut;

            // Add Endpoint address
            EndpointAddress rsEndpointAddress = new EndpointAddress(reportEndpoint);

            // Instantiate the Soap client
            ReportExecutionServiceSoapClient rsExec = new ReportExecutionServiceSoapClient(rsBinding, rsEndpointAddress);

            // Create a network credential object with the appropriate username and password used
            // to access the SSRS web service
            NetworkCredential clientCredentials = new NetworkCredential(reportEndpointUsername, reportEndpointPassword);

            // Here, for testing purposes, I use impersonation for a local account on the SSRS server
            //if (ConfigurationManager.AppSettings["ReportCredential"] == "WindowsCredential")
            //{
            rsExec.ClientCredentials.Windows.AllowedImpersonationLevel = System.Security.Principal.TokenImpersonationLevel.Impersonation;
            //}
            //else
            //{
            rsExec.ClientCredentials.Windows.ClientCredential = clientCredentials;
            //}

            string historyID = null;

            TrustedUserHeader trustedUserHeader = new TrustedUserHeader();
            ExecutionHeader execHeader = new ExecutionHeader();

            // Here we call the async LoadReport() method using the "await" keyword, which means any code below this method
            // will not execute until the result from the LoadReportAsync task is returned
            LoadReportResponse taskLoadReport = await rsExec.LoadReportAsync(trustedUserHeader, reportPath, historyID);
            // By the time the LoadReportAsync task is returned successfully, its "executionInfo" property
            // would have already been populated. Now the remaining code in this main thread will resume executing
            if (taskLoadReport.executionInfo != null)
            {
                execHeader.ExecutionID = taskLoadReport.executionInfo.ExecutionID;
            }

            await rsExec.SetExecutionParametersAsync(execHeader, trustedUserHeader, parameters, "en-us");

            byte[][] pages = new Byte[0][];
            string format = "IMAGE";
            int numberOfPages = 1;
            byte[] currentPageStream = new byte[1] { 0x00 }; // put a byte to get the loop started
            RenderRequest renderReq = null;
            RenderResponse taskRender = null;

            string dpiX = "200";
            string dpiY = "200";
            if(options != null)
            {
                dpiX = options.dpiX;
                dpiY = options.dpiY;
            }
            while (currentPageStream.Length > 0)
            {

                string deviceInfo = String.Format($@"<DeviceInfo><OutputFormat>EMF</OutputFormat><PrintDpiX>{dpiX}</PrintDpiX><PrintDpiY>{dpiY}</PrintDpiY>" + "<StartPage>{0}</StartPage></DeviceInfo>", numberOfPages);
                renderReq = new RenderRequest(execHeader, trustedUserHeader, format, deviceInfo);
                taskRender = await rsExec.RenderAsync(renderReq);
                currentPageStream = taskRender.Result;

                if (currentPageStream.Length == 0 && numberOfPages == 1)
                {
                    //renderException = EnumRenderException.ZERO_LENGTH_STREAM;
                    break;
                }

                //Add the byte stream of current page in pages[] array so that we can have complete report in pages[][] array
                if (currentPageStream.Length > 0)
                {
                    Array.Resize(ref pages, pages.Length + 1);
                    pages[pages.Length - 1] = currentPageStream;
                    numberOfPages++;
                }

            }

            ReportImageResult reportImageResult = new ReportImageResult { pages = pages, numberOfPages = numberOfPages - 1 };

            return reportImageResult;
        }

        private static async Task<byte[]> RenderReportAsync(string reportPath, string type, ParameterValue[] parameters)
        {
            string reportEndpoint = ConfigurationManager.AppSettings["ReportEndpoint"];
            string reportEndpointUsername = ConfigurationManager.AppSettings["ReportUsername"];
            string reportEndpointPassword = ConfigurationManager.AppSettings["ReportPassword"];

            reportPath = String.Concat("/PMS_Rep/", reportPath);
            TimeSpan SetTimeOut = new TimeSpan(0, 30, 0);
            BasicHttpBinding rsBinding = new BasicHttpBinding();
            rsBinding.Security.Mode = BasicHttpSecurityMode.TransportCredentialOnly;
            rsBinding.Security.Transport.ClientCredentialType = HttpClientCredentialType.Ntlm;
            rsBinding.MaxReceivedMessageSize = Int32.MaxValue;
            rsBinding.MaxBufferSize = Int32.MaxValue;
            rsBinding.OpenTimeout = SetTimeOut;
            rsBinding.CloseTimeout = SetTimeOut;
            rsBinding.SendTimeout = SetTimeOut;
            rsBinding.ReceiveTimeout = SetTimeOut;

            // Add Endpoint address
            EndpointAddress rsEndpointAddress = new EndpointAddress(reportEndpoint);

            // Instantiate the Soap client
            ReportExecutionServiceSoapClient rsExec = new ReportExecutionServiceSoapClient(rsBinding, rsEndpointAddress);

            // Create a network credential object with the appropriate username and password used
            // to access the SSRS web service
            NetworkCredential clientCredentials = new NetworkCredential(reportEndpointUsername, reportEndpointPassword);

            // Here, for testing purposes, I use impersonation for a local account on the SSRS server
            //if (ConfigurationManager.AppSettings["ReportCredential"] == "WindowsCredential")
            //{
            rsExec.ClientCredentials.Windows.AllowedImpersonationLevel = System.Security.Principal.TokenImpersonationLevel.Impersonation;
            //}
            //else
            //{
            rsExec.ClientCredentials.Windows.ClientCredential = clientCredentials;
            //}

            string historyID = null;

            TrustedUserHeader trustedUserHeader = new TrustedUserHeader();
            ExecutionHeader execHeader = new ExecutionHeader();

            // Here we call the async LoadReport() method using the "await" keyword, which means any code below this method
            // will not execute until the result from the LoadReportAsync task is returned
            LoadReportResponse taskLoadReport = await rsExec.LoadReportAsync(trustedUserHeader, reportPath, historyID);
            // By the time the LoadReportAsync task is returned successfully, its "executionInfo" property
            // would have already been populated. Now the remaining code in this main thread will resume executing
            if (taskLoadReport.executionInfo != null)
            {
                execHeader.ExecutionID = taskLoadReport.executionInfo.ExecutionID;
            }

            string deviceInfo = null;
            string format = type; //EXCELOPENXML, PDF

            await rsExec.SetExecutionParametersAsync(execHeader, trustedUserHeader, parameters, "en-us");

            // Now, similar to the above task, we will call the RenderAsync() method and await its result
            RenderRequest renderReq = new RenderRequest(execHeader, trustedUserHeader, format, deviceInfo);
            RenderResponse taskRender = await rsExec.RenderAsync(renderReq);
            // When the result is returned and if it is successfull, the rendered report's byte[] should be available

            return taskRender.Result;
        }

        private static void PushOutputFile(byte[] content, string outputPath)
        {
            //Create Directory If not exists
            FileInfo filePath = new FileInfo(outputPath);
            filePath.Directory.Create();

            //Save file to Directory
            File.WriteAllBytes(outputPath, content);
        }

    }
    public class ReportImageResult
    {
        public byte[][] pages { get; set; }
        public int numberOfPages { get; set; }
    }
    public class ReportOptions
    {
        public string dpiX { get; set; }
        public string dpiY { get; set; }
    }
}
